import logging
from ibmsecurity.utilities import tools
from ibmsecurity.isam.aac.api_protection import definitions

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/clients"
requires_modules = ["mga", "federation"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of API protection clients
    """
    return isamAppliance.invoke_get("Retrieve a list of API protection clients", uri, requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific API protection client
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    client_id = ret_obj['data']
    warnings = ret_obj["warnings"]

    if client_id == {}:
        logger.info("Client {0} had no match, skipping retrieval.".format(name))
        warnings.append("Client Name {0} had no match.".format(name))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, client_id)


def _get(isamAppliance, client_id):
    return isamAppliance.invoke_get("Retrieve a specific API protection client", "{0}/{1}".format(uri, client_id),
                                    requires_modules=requires_modules, requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search API Protection Client by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj['warnings'] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found API Protection Client {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj

def search_id(isamAppliance, clientId, check_mode=False, force=False):
    """
    Search API Protection Client by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj['warnings'] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['clientId'] == clientId:
            logger.info("Found API Protection Client {0} id: {1}".format(clientId, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj

def _get_id(isamAppliance, clientId, check_mode=False, force=False):
    """
    Search API Protection Client by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj['warnings'] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['clientId'] == clientId:
            logger.info("Found API Protection Client {0} id: {1}".format(clientId, obj['id']))
            return_obj['data'] = obj
            return_obj['rc'] = 0

    return return_obj


def generate_client_id(isamAppliance, check_mode=False, force=False):
    """
    Generate a client ID
    """
    return isamAppliance.invoke_get("Generate a client ID", "{0}/clientid".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def generate_client_secret(isamAppliance, check_mode=False, force=False):
    """
    Generate a client secret
    """
    return isamAppliance.invoke_get("Generate a client secret", "{0}/clientsecret".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def add(isamAppliance, name, definitionName, companyName, redirectUri=None, companyUrl=None, contactPerson=None,
        contactType=None, email=None, phone=None, otherInfo=None, clientId=None, clientSecret=None,
        requirePkce=None, encryptionDb=None, encryptionCert=None, jwksUri=None, extProperties=None, introspectWithSecret=None,
        check_mode=False, force=False):
    """
    Create an API protection definition
    """
    ret_obj = definitions.search(isamAppliance, definitionName, check_mode=check_mode, force=force)
    if ret_obj['data'] == {}:
        warnings = ret_obj["warnings"]
        warnings.append(
            "API Protection Definition {0} is not found. Cannot process client request.".format(definitionName))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        definition = ret_obj['data']

    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    warnings = ret_obj["warnings"]

    if force is True or ret_obj["data"] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the main client attributes
            client_json = {
                "name": name,
                "definition": definition,
                "companyName": companyName
            }
            # Add attributes that have been supplied... otherwise skip them.
            if redirectUri is not None:
                client_json["redirectUri"] = redirectUri
            if companyUrl is not None:
                client_json["companyUrl"] = companyUrl
            if contactPerson is not None:
                client_json["contactPerson"] = contactPerson
            if contactType is not None:
                client_json["contactType"] = contactType
            if email is not None:
                client_json["email"] = email
            if phone is not None:
                client_json["phone"] = phone
            if otherInfo is not None:
                client_json["otherInfo"] = otherInfo
            if clientId is not None:
                client_json["clientId"] = clientId
            if clientSecret is not None:
                client_json["clientSecret"] = clientSecret
            if requirePkce is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, requirePkce: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring requirePkce for this call.".format(
                            isamAppliance.facts["version"], requirePkce))
                else:
                    client_json["requirePkce"] = requirePkce
            if encryptionDb is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, encryptionDb: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionDb for this call.".format(
                            isamAppliance.facts["version"], encryptionDb))
                else:
                    client_json["encryptionDb"] = encryptionDb
            if encryptionCert is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, encryptionCert: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionCert for this call.".format(
                            isamAppliance.facts["version"], encryptionCert))
                else:
                    client_json["encryptionCert"] = encryptionCert
            if jwksUri is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, jwksUri: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring jwksUri for this call.".format(
                            isamAppliance.facts["version"], jwksUri))
                else:
                    client_json["jwksUri"] = jwksUri
            if extProperties is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, extProperties: {1} is not supported. Needs 9.0.5.0 or higher. Ignoring extProperties for this call.".format(
                            isamAppliance.facts["version"], extProperties))
                else:
                    client_json["extProperties"] = extProperties
            if introspectWithSecret is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, introspectWithSecret: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring introspectWithSecret for this call.".format(
                            isamAppliance.facts["version"], introspectWithSecret))
                else:
                    client_json["introspectWithSecret"] = introspectWithSecret

            return isamAppliance.invoke_post(
                "Create an API protection definition", uri, client_json, requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an API protection client registration
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    client_id = ret_obj['data']
    warnings = ret_obj["warnings"]

    if client_id == {}:
        logger.info("Client {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Delete an API protection client registration", "{0}/{1}".format(uri, client_id),
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, name, definitionName, companyName, redirectUri=None, companyUrl=None, contactPerson=None,
           contactType=None, email=None, phone=None, otherInfo=None, clientId=None, clientSecret=None,
           requirePkce=None, encryptionDb=None, encryptionCert=None, jwksUri=None, extProperties=None, introspectWithSecret=None,
           check_mode=False, force=False, new_name=None):
    """
    Update a specified mapping rule
    """
    ret_obj = definitions.search(isamAppliance, definitionName, check_mode=check_mode, force=force)
    if ret_obj['data'] == {}:
        warnings = ret_obj["warnings"]
        warnings.append(
            "API Protection Definition {0} is not found. Cannot process client request.".format(definitionName))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        definition = ret_obj['data']

    ret_obj = get(isamAppliance, name)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append("Client {0} not found, skipping update.".format(name))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        id = ret_obj["data"]["id"]

    needs_update = False
    # Create a simple json with just the main client attributes
    json_data = {
        "definition": definition,
        "companyName": companyName
    }
    if new_name is not None:
        json_data['name'] = new_name
    else:
        json_data['name'] = name

    del ret_obj['data']['id']
    # Add attributes that have been supplied... otherwise skip them.
    if redirectUri is not None:
        json_data["redirectUri"] = redirectUri
    elif 'redirectUri' in ret_obj['data']:
        if ret_obj['data']['redirectUri'] is not None:
            json_data['redirectUri'] = ret_obj['data']['redirectUri']
        else:
            del ret_obj['data']['redirectUri']
    if companyUrl is not None:
        json_data['companyUrl'] = companyUrl
    elif 'companyUrl' in ret_obj['data']:
        if ret_obj['data']['companyUrl'] is not None:
            json_data['companyUrl'] = ret_obj['data']['companyUrl']
        else:
            del ret_obj['data']['companyUrl']
    if contactPerson is not None:
        json_data['contactPerson'] = contactPerson
    elif 'contactPerson' in ret_obj['data']:
        if ret_obj['data']['contactPerson'] is not None:
            json_data['contactPerson'] = ret_obj['data']['contactPerson']
        else:
            del ret_obj['data']['contactPerson']
    if contactType is not None:
        json_data['contactType'] = contactType
    elif 'contactType' in ret_obj['data']:
        if ret_obj['data']['contactType'] is not None:
            json_data['contactType'] = ret_obj['data']['contactType']
        else:
            del ret_obj['data']['contactType']
    if email is not None:
        json_data['email'] = email
    elif 'email' in ret_obj['data']:
        if ret_obj['data']['email'] is not None:
            json_data['email'] = ret_obj['data']['email']
        else:
            del ret_obj['data']['email']
    if phone is not None:
        json_data['phone'] = phone
    elif 'phone' in ret_obj['data']:
        if ret_obj['data']['phone'] is not None:
            json_data['phone'] = ret_obj['data']['phone']
        else:
            del ret_obj['data']['phone']
    if otherInfo is not None:
        json_data['otherInfo'] = otherInfo
    elif 'otherInfo' in ret_obj['data']:
        if ret_obj['data']['otherInfo'] is not None:
            json_data['otherInfo'] = ret_obj['data']['otherInfo']
        else:
            del ret_obj['data']['otherInfo']
    if clientId is not None:
        json_data['clientId'] = clientId
    elif 'clientId' in ret_obj['data']:
        if ret_obj['data']['clientId'] is not None:
            json_data['clientId'] = ret_obj['data']['clientId']
        else:
            del ret_obj['data']['clientId']
    if clientSecret is not None:
        json_data['clientSecret'] = clientSecret
    elif 'clientSecret' in ret_obj['data']:
        if ret_obj['data']['clientSecret'] is not None:
            json_data['clientSecret'] = ret_obj['data']['clientSecret']
        else:
            del ret_obj['data']['clientSecret']
    if requirePkce is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, requirePkce: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring requirePkce for this call.".format(
                    isamAppliance.facts["version"], requirePkce))
        else:
            json_data['requirePkce'] = requirePkce
    elif 'requirePkce' in ret_obj['data']:
        if ret_obj['data']['requirePkce'] is not None:
            json_data['requirePkce'] = ret_obj['data']['requirePkce']
        else:
            del ret_obj['data']['requirePkce']
    if encryptionDb is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, encryptionDb: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionDb for this call.".format(
                    isamAppliance.facts["version"], encryptionDb))
        else:
            json_data['encryptionDb'] = encryptionDb
    elif 'encryptionDb' in ret_obj['data']:
        if ret_obj['data']['encryptionDb'] is not None:
            json_data['encryptionDb'] = ret_obj['data']['encryptionDb']
        else:
            del ret_obj['data']['encryptionDb']
    if encryptionCert is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, encryptionCert: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionCert for this call.".format(
                    isamAppliance.facts["version"], encryptionCert))
        else:
            json_data['encryptionCert'] = encryptionCert
    elif 'encryptionCert' in ret_obj['data']:
        if ret_obj['data']['encryptionCert'] is not None:
            json_data['encryptionCert'] = ret_obj['data']['encryptionCert']
        else:
            del ret_obj['data']['encryptionCert']
    if jwksUri is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, jwksUri: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring jwksUri for this call.".format(
                    isamAppliance.facts["version"], jwksUri))
        else:
            json_data['jwksUri'] = jwksUri
    elif 'jwksUri' in ret_obj['data']:
        if ret_obj['data']['jwksUri'] is not None:
            json_data['jwksUri'] = ret_obj['data']['jwksUri']
        else:
            del ret_obj['data']['jwksUri']
    if extProperties is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
            warnings.append(
                "Appliance at version: {0}, extProperties: {1} is not supported. Needs 9.0.5.0 or higher. Ignoring extProperties for this call.".format(
                    isamAppliance.facts["version"], extProperties))
        else:
            json_data['extProperties'] = extProperties
    elif 'extProperties' in ret_obj['data']:
        if ret_obj['data']['extProperties'] is not None:
            json_data['extProperties'] = ret_obj['data']['extProperties']
        else:
            del ret_obj['data']['extProperties']
    if introspectWithSecret is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
            warnings.append(
                "Appliance at version: {0}, introspectWithSecret: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring introspectWithSecret for this call.".format(
                    isamAppliance.facts["version"], introspectWithSecret))
        else:
            json_data["introspectWithSecret"] = introspectWithSecret
    elif 'introspectWithSecret' in ret_obj['data']:
        if ret_obj['data']['introspectWithSecret'] is not None:
            json_data['introspectWithSecret'] = ret_obj['data']['introspectWithSecret']
        else:
            del ret_obj['data']['introspectWithSecret']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))
    if sorted_ret_obj != sorted_json_data:
        needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update a specified mapping rule", "{0}/{1}".format(uri, id), json_data,
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, name, definitionName, companyName, redirectUri=None, companyUrl=None, contactPerson=None,
        contactType=None, email=None, phone=None, otherInfo=None, clientId=None, clientSecret=None, new_name=None,
        requirePkce=None, encryptionDb=None, encryptionCert=None, jwksUri=None, extProperties=None, introspectWithSecret=None,
        check_mode=False, force=False):
    """
    Creating or Modifying an API Protection Definition
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know policy does not exist
        logger.info("Definition {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name, definitionName, companyName, redirectUri=redirectUri, companyUrl=companyUrl,
                   contactPerson=contactPerson, contactType=contactType, email=email, phone=phone, otherInfo=otherInfo,
                   clientId=clientId, clientSecret=clientSecret, requirePkce=requirePkce, encryptionDb=encryptionDb,
                   encryptionCert=encryptionCert, jwksUri=jwksUri, extProperties=extProperties, introspectWithSecret=introspectWithSecret,
                   check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Definition {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name, definitionName, companyName, redirectUri=redirectUri, companyUrl=companyUrl,
                      contactPerson=contactPerson, contactType=contactType, email=email, phone=phone,
                      otherInfo=otherInfo, clientId=clientId, clientSecret=clientSecret, new_name=new_name,
                      requirePkce=requirePkce, encryptionDb=encryptionDb, encryptionCert=encryptionCert,
                      jwksUri=jwksUri, extProperties=extProperties, introspectWithSecret=introspectWithSecret, check_mode=check_mode, force=force)

def update_id(isamAppliance, clientId, name, definitionName, companyName, redirectUri=None, companyUrl=None, contactPerson=None,
           contactType=None, email=None, phone=None, otherInfo=None, clientSecret=None,
           requirePkce=None, encryptionDb=None, encryptionCert=None, jwksUri=None, extProperties=None, introspectWithSecret=None,
           check_mode=False, force=False, new_clientId=None):
    """
    Update a specified mapping rule
    """

    ret_obj = definitions.search(isamAppliance, definitionName, check_mode=check_mode, force=force)
    if ret_obj['data'] == {}:
        warnings = ret_obj["warnings"]
        warnings.append(
            "API Protection Definition {0} is not found. Cannot process client request.".format(definitionName))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        definition = ret_obj['data']

    ret_obj = _get_id(isamAppliance, clientId=clientId)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append("Client {0} not found, skipping update.".format(clientId))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        id = ret_obj["data"]["id"]

    needs_update = False
    # Create a simple json with just the main client attributes
    json_data = {
        "definition": definition,
        "companyName": companyName,
        "name": name
    }

    if new_clientId is not None:
        json_data['clientId'] = new_clientId
    else:
        json_data['clientId'] = clientId

    del ret_obj['data']['id']
    # Add attributes that have been supplied... otherwise skip them.
    if redirectUri is not None:
        json_data['redirectUri'] = redirectUri
    elif 'redirectUri' in ret_obj['data']:
        if ret_obj['data']['redirectUri'] is not None:
            json_data["redirectUri"] = ret_obj['data']['redirectUri']
        else:
            del ret_obj['data']['redirectUri']
    if companyUrl is not None:
        json_data["companyUrl"] = companyUrl
    elif 'companyUrl' in ret_obj['data']:
        if ret_obj['data']['companyUrl'] is not None:
            json_data["companyUrl"] = ret_obj['data']['companyUrl']
        else:
            del ret_obj['data']['companyUrl']
    if contactPerson is not None:
        json_data["contactPerson"] = contactPerson
    elif 'contactPerson' in ret_obj['data']:
        if ret_obj['data']['contactPerson'] is not None:
            json_data["contactPerson"] = ret_obj['data']['contactPerson']
        else:
            del ret_obj['data']['contactPerson']
    if contactType is not None:
        json_data["contactType"] = contactType
    elif 'contactType' in ret_obj['data']:
        if ret_obj['data']['contactType'] is not None:
            json_data["contactType"] = ret_obj['data']['contactType']
        else:
            del ret_obj['data']['contactType']
    if email is not None:
        json_data["email"] = email
    elif 'email' in ret_obj['data']:
        if ret_obj['data']['email'] is not None:
            json_data["email"] = ret_obj['data']['email']
        else:
            del ret_obj['data']['email']
    if phone is not None:
        json_data["phone"] = phone
    elif 'phone' in ret_obj['data']:
        if ret_obj['data']['phone'] is not None:
            json_data["phone"] = ret_obj['data']['phone']
        else:
            del ret_obj['data']['phone']
    if otherInfo is not None:
        json_data["otherInfo"] = otherInfo
    elif 'otherInfo' in ret_obj['data']:
        if ret_obj['data']['otherInfo'] is not None:
            json_data["otherInfo"] = ret_obj['data']['otherInfo']
        else:
            del ret_obj['data']['otherInfo']
    if clientSecret is not None:
        json_data["clientSecret"] = clientSecret
    elif 'clientSecret' in ret_obj['data']:
        if ret_obj['data']['clientSecret'] is not None:
            json_data["clientSecret"] = ret_obj['data']['clientSecret']
        else:
            del ret_obj['data']['clientSecret']
    if requirePkce is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, requirePkce: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring requirePkce for this call.".format(
                    isamAppliance.facts["version"], requirePkce))
        else:
            json_data["requirePkce"] = requirePkce
    elif 'requirePkce' in ret_obj['data']:
        if ret_obj['data']['requirePkce'] is not None:
            json_data["requirePkce"] = ret_obj['data']['requirePkce']
        else:
            del ret_obj['data']['requirePkce']
    if encryptionDb is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, encryptionDb: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionDb for this call.".format(
                    isamAppliance.facts["version"], encryptionDb))
        else:
            json_data["encryptionDb"] = encryptionDb
    elif 'encryptionDb' in ret_obj['data']:
        if ret_obj['data']['encryptionDb'] is not None:
            json_data["encryptionDb"] = ret_obj['data']['encryptionDb']
        else:
            del ret_obj['data']['encryptionDb']
    if encryptionCert is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, encryptionCert: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring encryptionCert for this call.".format(
                    isamAppliance.facts["version"], encryptionCert))
        else:
            json_data["encryptionCert"] = encryptionCert
    elif 'encryptionCert' in ret_obj['data']:
        if ret_obj['data']['encryptionCert'] is not None:
            json_data["encryptionCert"] = ret_obj['data']['encryptionCert']
        else:
            del ret_obj['data']['encryptionCert']
    if jwksUri is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, jwksUri: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring jwksUri for this call.".format(
                    isamAppliance.facts["version"], jwksUri))
        else:
            json_data["jwksUri"] = jwksUri
    elif 'jwksUri' in ret_obj['data']:
        if ret_obj['data']['jwksUri'] is not None:
            json_data["jwksUri"] = ret_obj['data']['jwksUri']
        else:
            del ret_obj['data']['jwksUri']
    if extProperties is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
            warnings.append(
                "Appliance at version: {0}, extProperties: {1} is not supported. Needs 9.0.5.0 or higher. Ignoring extProperties for this call.".format(
                    isamAppliance.facts["version"], extProperties))
        else:
            json_data["extProperties"] = extProperties
    elif 'extProperties' in ret_obj['data']:
        if ret_obj['data']['extProperties'] is not None:
            json_data["extProperties"] = ret_obj['data']['extProperties']
        else:
            del ret_obj['data']['extProperties']
    if introspectWithSecret is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
            warnings.append(
                "Appliance at version: {0}, introspectWithSecret: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring introspectWithSecret for this call.".format(
                    isamAppliance.facts["version"], introspectWithSecret))
        else:
            json_data["introspectWithSecret"] = introspectWithSecret
    elif 'introspectWithSecret' in ret_obj['data']:
        if ret_obj['data']['introspectWithSecret'] is not None:
            json_data["introspectWithSecret"] = ret_obj['data']['introspectWithSecret']
        else:
            del ret_obj['data']['introspectWithSecret']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))
    if sorted_ret_obj != sorted_json_data:
        needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update a specified mapping rule", "{0}/{1}".format(uri, id), json_data,
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def set_id(isamAppliance, clientId, name, definitionName, companyName, redirectUri=None, companyUrl=None, contactPerson=None,
        contactType=None, email=None, phone=None, otherInfo=None, clientSecret=None, new_clientId=None,
        requirePkce=None, encryptionDb=None, encryptionCert=None, jwksUri=None, extProperties=None, introspectWithSecret=None,
        check_mode=False, force=False):
    """
    Creating or Modifying an API Protection Definition
    """
    if (search_id(isamAppliance, clientId))['data'] == {}:
        # Force the add - we already know policy does not exist
        logger.info("Definition {0} had no match, requesting to add new one.".format(clientId))
        return add(isamAppliance, name=name, definitionName=definitionName, companyName=companyName, redirectUri=redirectUri, companyUrl=companyUrl,
                   contactPerson=contactPerson, contactType=contactType, email=email, phone=phone, otherInfo=otherInfo,
                   clientId=clientId, clientSecret=clientSecret, requirePkce=requirePkce, encryptionDb=encryptionDb,
                   encryptionCert=encryptionCert, jwksUri=jwksUri, extProperties=extProperties, introspectWithSecret=introspectWithSecret,
                   check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Definition {0} exists, requesting to update.".format(clientId))
        return update_id(isamAppliance, clientId=clientId, definitionName=definitionName, companyName=companyName, name=name, redirectUri=redirectUri, companyUrl=companyUrl,
                      contactPerson=contactPerson, contactType=contactType, email=email, phone=phone,
                      otherInfo=otherInfo, clientSecret=clientSecret, new_clientId=new_clientId,
                      requirePkce=requirePkce, encryptionDb=encryptionDb, encryptionCert=encryptionCert,
                      jwksUri=jwksUri, extProperties=extProperties, introspectWithSecret=introspectWithSecret, check_mode=check_mode, force=force)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare API Protection Definitions between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['definition']
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['definition']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'definition'])



import logging
import ibmsecurity.utilities.tools
from datetime import datetime
import json

performCertCheck = True
try:
    from dateutil import parser
    #pip install python-dateutil
    from OpenSSL.crypto import load_pkcs12, dump_certificate, load_certificate, FILETYPE_PEM
    # pip install pyOpenSSL
except:
    performCertCheck = False

logger = logging.getLogger(__name__)

def get(isamAppliance, check_mode=False, force=False):
    """
    Get management ssl certificate information
    """
    return isamAppliance.invoke_get("Get management ssl certificate information",
                                    "/isam/management_ssl_certificate/")


def set(isamAppliance, certificate, password, check_mode=False, force=False):
    """
    Import certificate database
    """
    if not performCertCheck:
        warnings = ["Idempotency not available. Unable to extract existing certificate to compare with provided one.  Install Python modules python-dateutil and pyOpenSSL."]
    else:
        warnings = None
    if force is True or not _check(isamAppliance, certificate, password):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post_files(
                "Import certificate database",
                "/isam/management_ssl_certificate",
                [
                    {
                        'file_formfield': 'cert',
                        'filename': certificate,
                        'mimetype': 'application/x-pkcs12'
                    }
                ],
                {
                    'password': password
                },
                json_response=False, warnings=warnings)

    return isamAppliance.create_return_object()


def _check(isamAppliance, certificate, password):
    """
    requires additionally pyOpenSSL to load the p12 and extract the issuer, subject, etc
    Comparing issuer, subject, notafter (date only) and notbefore (date only)
    This DOES NOT check if the certificate is newer , just that it's different from the one that is deployed.
    """
    if performCertCheck:
        currentCert = get(isamAppliance)
        currentCert = currentCert['data']

        # remove values that are not checked
        del currentCert['keysize']
        del currentCert['version']

        # use date only to compare.  this simplifies the logic here, Python is not very strong with comparing dates in different timezones
        currentCert['notbefore'] = parser.parse(currentCert['notbefore'], ignoretz=True).strftime("%Y-%m-%d")
        currentCert['notafter'] = parser.parse(currentCert['notafter'], ignoretz=True).strftime("%Y-%m-%d")

        newCert = {}
        _type = FILETYPE_PEM
        _enc = 'utf-8'
        with open(certificate, 'rb') as f:
            c = f.read()
        p = load_pkcs12(c, password)

        certificate = p.get_certificate()
        newCert['subject'] = ",".join(f"{str(name, 'utf-8')}={str(value, 'utf-8')}" for name, value in reversed(certificate.get_subject().get_components()))

        x509 = load_certificate(_type, dump_certificate(_type, certificate))

        newCert['issuer'] = ",".join(f"{str(name, _enc)}={str(value, _enc)}" for name, value in reversed(x509.get_issuer().get_components()))
        newCert['notafter'] = datetime.strptime(str(x509.get_notAfter(), _enc), "%Y%m%d%H%M%SZ").strftime("%Y-%m-%d")
        newCert['notbefore'] = datetime.strptime(str(x509.get_notBefore(), _enc), "%Y%m%d%H%M%SZ").strftime("%Y-%m-%d")

        curc = json.dumps(currentCert, skipkeys=True, sort_keys=True)
        logger.debug(f"\nSorted Current  Management Cert:\n {curc}\n")

        newc = json.dumps(newCert, skipkeys=True, sort_keys=True)
        logger.debug(f"\nSorted Desired  Management Cert:\n {newc}\n")

        if curc == newc:
            return True
        else:
            return False
    else:
        logger.info('Skipping management certificate check because pyOpenSSL or not available.  Install with pip install pyOpenSSL')
        return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare management authentication settings
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve available updates
    """
    return isvgAppliance.invoke_get("Retrieving available updates",
                                    "/updates/available.json")


def discover(isvgAppliance, check_mode=False, force=False):
    """
    Discover available updates
    """
    return isvgAppliance.invoke_get("Discover available updates",
                                    "/updates/available/discover")


def upload(isvgAppliance, file, check_mode=False, force=False):
    """
    Upload Available Update
    """
    if force is True or _check_file(isvgAppliance, file) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post_files(
                "Upload Available Update",
                "/updates/available",
                [{
                    'file_formfield': 'uploadedfile',
                    'filename': file,
                    'mimetype': 'application/octet-stream'
                }],
                {}, json_response=False)

    return isvgAppliance.create_return_object()


def _check_file(isvgAppliance, file):
    """
    Parse the file name to see if it is already uploaded - use version and release date from pkg file name
    Also check to see if the firmware level is already uploaded
    Note: Lot depends on the name of the file.

    :param isvgAppliance:
    :param file:
    :return:
    """
    import os.path

    # If there is an exception then simply return False
    # Sample filename - 8.0.1.9-ISS-ISVG_20181207-0045.pkg
    logger.debug("Checking provided file is ready to upload: {0}".format(file))
    try:
        # Extract file name from path
        f = os.path.basename(file)
        fn = os.path.splitext(f)
        logger.debug("File name without path: {0}".format(fn[0]))

        # Split of file by '-' hyphen and '_' under score
        import re
        fp = re.split('-|_', fn[0])
        firm_file_version = fp[0]
        firm_file_product = fp[2]
        firm_file_date = fp[3]
        logger.debug("PKG file details: {0}: version: {1} date: {2}".format(firm_file_product, firm_file_version, firm_file_date))

        # Check if firmware level already contains the update to be uploaded or greater, check Active partition
        # firmware "name" of format - 8.0.1.9-ISS-ISVG_20181207-0045
        import ibmsecurity.isvg.firmware
        ret_obj = ibmsecurity.isvg.firmware.get(isvgAppliance)
        for firm in ret_obj['data']:
            # Split of file by '-' hyphen and '_' under score
            fp = re.split('-|_', firm['name'])
            firm_appl_version = fp[0]
            firm_appl_product = fp[2]
            firm_appl_date = fp[3]
            logger.debug("Partition details ({0}): {1}: version: {2} date: {3}".format(firm['partition'], firm_appl_product, firm_appl_version, firm_appl_date))
            if firm['active'] is True:
                from ibmsecurity.utilities import tools
                if tools.version_compare(firm_appl_version, firm_file_version) >= 0:
                    logger.info(
                        "Active partition has version {0} which is greater or equals than install package at version {1}.".format(
                            firm_appl_version, firm_file_version))
                    return True
                else:
                    logger.info(
                        "Active partition has version {0} which is smaller than install package at version {1}.".format(
                            firm_appl_version, firm_file_version))

        # Check if update uploaded - will not show up if installed though
        ret_obj = get(isvgAppliance)
        for upd in ret_obj['data']:
            rd = upd['release_date']
            rd = rd.replace('-', '')  # turn release date into 20161102 format from 2016-11-02
            if upd['version'] == fp[0] and rd == fp[3]:  # Version of format 8.0.1.9
                return True
    except Exception as e:
        logger.debug("Exception occured: {0}".format(e))
        pass

    return False


def install(isvgAppliance, type, version, release_date, name, check_mode=False, force=False):
    """
    Install Available Update
    """
    if force is True or _check(isvgAppliance, type, version, release_date, name) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            # obtain appliance lastboot time
            import ibmsecurity.isvg.firmware
            import time
            ret_obj_appliance = ibmsecurity.isvg.firmware.get(isvgAppliance)
            for firm in ret_obj_appliance['data']:
                if firm['active'] is True:
                    before_install_last_boot = firm['last_boot']
                    logger.info(
                        "Active partition last boot time {0} before firmware installation.".format(before_install_last_boot))
            
            # proceed with firmware installation
            ret_obj = isvgAppliance.invoke_post("Install Available Update",
                                                "/updates/available/install",
                                                {"updates": [
                                                    {
                                                        "type": type,
                                                        "version": version,
                                                        "release_date": release_date,
                                                        "name": name
                                                    }
                                                ]
                                                })
            isvgAppliance.facts['version'] = version
            
            # isvg appliance has so much work to do after installing the firmware, it can take up
            # to several minutes before it reboots by itself.
            # Loop until appliance returns an error, or reboot is detected, before returning control.
            try:
                while True:
                    ret_obj_appliance = ibmsecurity.isvg.firmware.get(isvgAppliance)
                    for firm in ret_obj_appliance['data']:
                        if firm['active'] is True:
                            last_boot = firm['last_boot']
                            logger.info(
                                "Active partition last boot time {0} after firmware installation.".format(last_boot))
                            if last_boot > before_install_last_boot:
                                raise Exception ("Break out of loop")
                    time.sleep(30)
            except Exception as e:
                logger.debug("Exception occured: {0}. Assuming appliance has now initiated reboot process".format(e))
                pass
            
            return ret_obj

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, type, version, release_date, name):
    ret_obj = get(isvgAppliance)
    for upd in ret_obj['data']:
        # If there is an installation in progress then abort
        if upd['state'] == 'Installing':
            logger.debug("Detecting a state of installing...")
            return False
        if upd['type'] == type and upd['version'] == version and upd['release_date'] == release_date and upd[
            'name'] == name:
            logger.debug("Requested firmware ready for install...")
            return True

    logger.debug("Requested firmware not available for install...")

    return False



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/piptypes"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, count=None, start=None, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of policy information point types
    """
    return isamAppliance.invoke_get("Retrieve a list of policy information point types",
                                    "{0}/{1}".format(uri,
                                                     tools.create_query_string(count=count, start=start, filter=filter,
                                                                               sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific policy information point type
    """
    ret_obj = search(isamAppliance, name, check_mode=False, force=False)
    id = ret_obj['data']

    if id == {}:
        logger.info("PIP Type {0} had no match, skipping retrieval.".format(name))
        warnings = ["PIP Type {0} had no match, skipping retrieval.".format(name)]
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, id)


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Retrieve ID for named PIP
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Policy Type {0} id: {1}".format(name, obj['id']))
            ret_obj_new['data'] = obj['id']

    return ret_obj_new


def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific PIP Type",
                                    "{0}/{1}".format(uri, id))



import logging
import ibmsecurity.isam.aac.server_connections.ws
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"

def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current enterprise user profile SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current enterprise user profile SCIM configuration settings",
                                    "{0}/urn:ietf:params:scim:schemas:extension:enterprise:2.0:User".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)

def set(isamAppliance, mappings, check_mode=False, force=False):
    """
    Updating the enterprise user profile SCIM configuration settings

    """

    ret_obj = get(isamAppliance)
    old_obj = ret_obj['data']['urn:ietf:params:scim:schemas:extension:enterprise:2.0:User']['mappings']
    new_obj = mappings['mappings']

    old_obj = json_sort(old_obj)
    new_obj = json_sort(new_obj)

    update_required = False
    found = False

    if len(old_obj) >= 1:
        if len(new_obj) >= 1:
            for obj in new_obj:
                for maping in ret_obj:
                    if obj == maping:
                        found = True
                if found is False:
                    update_required = True
        else:
            update_required = True
    else:
        if len(new_obj) >= 1:
            update_required = True


    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Updating the enterprise user profile SCIM configuration settings",
                                    "{0}/urn:ietf:params:scim:schemas:extension:enterprise:2.0:User".format(uri),
                                    mappings,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )

    return isamAppliance.create_return_object(changed=False)


import logging

logger = logging.getLogger(__name__)
uri = "/iam/access/v8/mmfa-config"
requires_modules = None
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve MMFA endpoint details
    """
    return isamAppliance.invoke_get("Retrieve MMFA endpoint details",
                                    "/iam/access/v8/mmfa-config")


def set(isamAppliance, client_id, options, endpoints, discovery_mechanisms, check_mode=False, force=False):
    """
    Set MMFA endpoint configuration
    """
    json_data = {
        "client_id": client_id,
        "options": options,
        "endpoints": endpoints,
        "discovery_mechanisms": discovery_mechanisms
    }
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post(
            "Set MMFA endpoint configuration",
            "/iam/access/v8/mmfa-config/", json_data)


def delete(isamAppliance, check_mode=False, force=False):
    """
    Unconfigure MMFA endpoint details
    """
    ret_obj = get(isamAppliance)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Unconfigure MMFA endpoint details",
                "{0}".format(uri),
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Get management authorization - config
    """
    return isamAppliance.invoke_get("Get management authorization - config",
                                    "/authorization/v1")


def _check(isamAppliance):
    """
    Check if management authorization is enabled
    """
    ret_obj = get(isamAppliance)

    return ret_obj['data']['config']['enforcing']


def set(isamAppliance, enforcing=True, check_mode=False, force=False):
    """
    Set remote management authorization
    """
    if force is True or _check(isamAppliance) != enforcing:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Set management authorization - config",
                "/authorization/config/v1",
                {
                    'enforcing': enforcing
                })

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare management authorization config settings
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/domain_realm/"
requires_modules = ['wga']
requires_version = None


def get_all(isamAppliance, recursive='yes', includeValues='yes', check_mode=False, force=False):
    """
    Retrieve Kerberos Configuration: all Domain to Realm mappings defined
    """
    return isamAppliance.invoke_get("Retrieve Kerberos Configuration: Domains",
                                    "{0}{1}".format(uri, tools.create_query_string(recursive=recursive,
                                                                                   includeValues=includeValues)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, name, value, check_mode=False, force=False):
    """
    Creating a Kerberos Doamin (Local DNS Value)
    """
    ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a Kerberos domain",
                "{0}".format(uri),
                {
                    "name": name,
                    "value": value
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Deleting a Kerberos Doamin
    """
    ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a Kerberos realm",
                "{0}{1}".format(uri, name))

    return isamAppliance.create_return_object()


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Get a specific kerberos domain by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Kerberos domain {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj
            return_obj['rc'] = 0

    return return_obj


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search kerberos domain to realm mapping by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Kerberos domain {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def update(isamAppliance, name, value, check_mode=False, force=False):
    """
    Update a specified API domain mapping
    """
    warnings = []
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        warnings.append("Name: {0} not found, hence cannot update.".format(name))
    elif ret_obj['data']['value'] != value or force is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified domain to realm mapping entry",
                "{0}{1}".format(uri, name), {
                    "id": name,
                    "value": value
                }, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)

def set(isamAppliance, name, value, check_mode=False, force=False):
    """
    Set a specified Kerberos domain mapping

    :param isamAppliance:
    :return:
    """
    ret_obj = search(isamAppliance, name)

    if ret_obj["data"] == {}:
        return add(isamAppliance, name, value, check_mode, force=True)

    return update(isamAppliance,  name, value, check_mode, force)

def compare(isamAppliance1, isamAppliance2):
    """
    Compare kerberos configuration domains between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, check_mode=False, force=False):
    """
    Retrieve existing firmware.
    """
    return isdsAppliance.invoke_get("Retrieving firmware",
                                    "/firmware_settings")


def backup(isdsAppliance, check_mode=False, force=False):
    """
    Kickoff Backup of active partition
    """
    if check_mode is True:
        return isdsAppliance.create_return_object(changed=True)
    else:
        return isdsAppliance.invoke_put("Kickoff Backup of Active Partition",
                                        "/firmware_settings/kickoff_backup", {})


def swap(isdsAppliance, check_mode=False, force=False):
    """
    Kickoff swap of active partition
    """
    if check_mode is True:
        return isdsAppliance.create_return_object(changed=True)
    else:
        ret_obj_old = get(isdsAppliance)

        ret_obj = isdsAppliance.invoke_put("Kickoff swap of Active Partition",
                                           "/firmware_settings/kickoff_swap", {})
        # Process previous query after a successful call to swap the partition
        for partition in ret_obj_old['data']:
            if partition['active'] is False:  # Get version of inactive partition (active now!)
                ver = partition['firmware_version'].split(' ')
                isdsAppliance.facts['version'] = ver[-1]

        return ret_obj


def set(isdsAppliance, id, comment, check_mode=False, force=False):
    """
    Update comment on partition
    """
    if force is True or _check_comment(isdsAppliance, comment) is False:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_put("Update comment on Partition",
                                            "/firmware_settings/{0}".format(id),
                                            {'comment': comment})

    return isdsAppliance.create_return_object()


def _check_comment(isdsAppliance, comment):
    ret_obj = get(isdsAppliance)

    # Loop through firmware partitions looking for active one
    for partition in ret_obj['data']:
        if partition['active'] is True:
            return (partition['comment'] == comment)

    return False


def compare(isdsAppliance1, isdsAppliance2):
    """
    Compare firmware between two appliances
    """
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    for obj in ret_obj1['data']:
        del obj['install_date']
        del obj['backup_date']
        del obj['last_boot']
    for obj in ret_obj2['data']:
        del obj['install_date']
        del obj['backup_date']
        del obj['last_boot']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2,
                                                    deleted_keys=['install_date', 'backup_date', 'last_boot'])



import logging
import ibmsecurity.utilities.tools
import os.path

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get information on existing snapshots
    """
    return isvgAppliance.invoke_get("Retrieving snapshots", "/snapshots")


def create(isvgAppliance, comment='', check_mode=False, force=False):
    """
    Create a new snapshot
    """
    if force is True or _check(isvgAppliance, comment=comment) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post("Creating snapshot", "/snapshots",
                                             {
                                                 'comment': comment
                                             })

    return isvgAppliance.create_return_object()


def delete(isvgAppliance, id, check_mode=False, force=False):
    """
    Delete a snapshot
    """
    if force is True or _check(isvgAppliance, id=id) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_delete("Deleting snapshot", "/snapshots/" + id)

    return isvgAppliance.create_return_object()


def modify(isvgAppliance, id, comment, check_mode=False, force=False):
    """
    Modify the snapshot comment
    """
    if force is True or _check(isvgAppliance, id=id) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_put("Modifying snapshot", "/snapshots/" + id,
                                            {
                                                'comment': comment
                                            })

    return isvgAppliance.create_return_object()


def apply(isvgAppliance, id, check_mode=False, force=False):
    """
    Apply a snapshot
    """
    uri = "/snapshots/apply/" + id
    if force is True or _check(isvgAppliance, id=id) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post("Applying snapshot", uri,
                                             {"authenticity_token": "KRVlgWbaS4GCOxfZcw+hkrPH2fVB1Vi+k8cayf0y9T4"})

    return isvgAppliance.create_return_object()


def download(isvgAppliance, filename, id, check_mode=False, force=False):
    """
    Download one snapshot file to a zip file.
    TODO: Can handle multiple id's - but rest of logic deals with just one for now
    """
    if force is True or (_check(isvgAppliance, id=id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            uri = "/snapshots/download?record_ids={0}".format(id)
            return isvgAppliance.invoke_get_file("Downloading snapshots", uri, filename)

    return isvgAppliance.create_return_object()


def download_latest(isvgAppliance, dir='.', check_mode=False, force=False):
    """
    Download latest snapshot file to a zip file.
    """
    ret_obj = get(isvgAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    snaps = min(ret_obj['data'], key=lambda snap: snap['index'])
    id = snaps['id']
    file = snaps['filename']
    filename = os.path.join(dir, file)

    return download(isvgAppliance, filename, id, check_mode, force)


def upload(isvgAppliance, filename, check_mode=False, force=False):
    """
    Upload snapshot
    """
    uri = "/snapshots/upload"
    if _check(isvgAppliance, fn=filename) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post_files(
                description="Upload snapshot",
                uri="{0}".format(uri),
                fileinfo=[{
                    'file_formfield': 'file',
                    'filename': filename,
                    'mimetype': 'application/octet-stream'
                }],
                data={},
                json_response=False)

    return isvgAppliance.create_return_object()


def apply_latest(isvgAppliance, check_mode=False, force=False):
    """
    Apply latest snapshot file (revert to latest)
    """
    ret_obj = get(isvgAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    snaps = min(ret_obj['data'], key=lambda snap: snap['index'])
    id = snaps['id']

    return apply(isvgAppliance, id, check_mode, force)


def compare(isvgAppliance1, isvgAppliance2):
    """
    Compare list of snapshots between 2 appliances
    """
    ret_obj1 = get(isvgAppliance1)
    ret_obj2 = get(isvgAppliance2)

    # id of snapshot is uniquely generated on appliance and should therefore be ignored in comparison.
    # filename of snapshot is generated based on exact date/time and will differ even if 2 snapshots are taken near the
    # same time. Therefore, filename should be ignored in comparison
    for snapshot in ret_obj1['data']:
        del snapshot['id']
        del snapshot['filename']

    for snapshot in ret_obj2['data']:
        del snapshot['id']
        del snapshot['filename']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'filename'])


def _check(isvgAppliance, comment='', id=None, fn=None):
    """
    Check if the last created snapshot has the exact same comment or id exists

    :param isvgAppliance:
    :param comment:
    :return:
    """
    ret_obj = get(isvgAppliance)

    if id != None:
        for snaps in ret_obj['data']:
            if snaps['id'] == id:
                return True
    elif fn != None:
        for snaps in ret_obj['data']:
            if snaps['filename'] == fn:
                return True
    else:
        for snaps in ret_obj['data']:
            if comment in snaps['comment']:
                return True

    return False



import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/cors"
requires_modules = ["wga"]
requires_version = "10.0.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of all existing CORS Policies
    """
    return isamAppliance.invoke_get("Retrieving the list of all existing CORS Policies",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, cors_policy_name, check_mode=False, force=False):
    """
    Retrieve a single CORS Policy
    """
    return isamAppliance.invoke_get("Retrieve a single CORS Policy",
                                    "{0}/{1}".format(uri, cors_policy_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, name, allowed_origins, allow_credentials=False, exposed_headers=[], handle_preflight=False,
        allowed_methods=[], allowed_headers=[], max_age=0, check_mode=False, force=False):
    """
    Create a new API Access Control Policy
    """
    exist, warnings = _check_exist(isamAppliance, name)

    if force is True or exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new API Access Control Policy", uri,
                {
                    "name": name,
                    "allowed_origins": allowed_origins,
                    "allow_credentials": allow_credentials,
                    "exposed_headers": exposed_headers,
                    "handle_preflight": handle_preflight,
                    "allowed_methods": allowed_methods,
                    "allowed_headers": allowed_headers,
                    "max_age": max_age
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, cors_policy_name, allowed_origins, allow_credentials=False, exposed_headers=[],
           handle_preflight=False, allowed_methods=[], allowed_headers=[], max_age=0, check_mode=False, force=False):
    """
    Update an existing API Access Control Policy
    """
    update_required, warnings, json_data = _check(isamAppliance=isamAppliance, cors_policy_name=cors_policy_name,
                                                  allowed_origins=allowed_origins, allow_credentials=allow_credentials,
                                                  exposed_headers=exposed_headers, handle_preflight=handle_preflight,
                                                  allowed_methods=allowed_methods, allowed_headers=allowed_headers,
                                                  max_age=max_age)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update an existing API Access Control Policy",
                "{0}/{1}".format(uri, cors_policy_name),
                json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, cors_policy_name, allowed_origins, allow_credentials=False, exposed_headers=[],
        handle_preflight=False, allowed_methods=[], allowed_headers=[], max_age=0, check_mode=False, force=False):
    exist, warnings = _check_exist(isamAppliance, cors_policy_name)
    if exist:
        return update(isamAppliance=isamAppliance, cors_policy_name=cors_policy_name, allowed_origins=allowed_origins,
                      allow_credentials=allow_credentials, exposed_headers=exposed_headers,
                      handle_preflight=handle_preflight, allowed_methods=allowed_methods,
                      allowed_headers=allowed_headers, max_age=max_age)
    else:
        return add(isamAppliance=isamAppliance, name=cors_policy_name, allowed_origins=allowed_origins,
                   allow_credentials=allow_credentials, exposed_headers=exposed_headers,
                   handle_preflight=handle_preflight, allowed_methods=allowed_methods, allowed_headers=allowed_headers,
                   max_age=max_age)


def delete(isamAppliance, cors_policy_name, check_mode=False, force=False):
    """
    Delete an access policy
    """
    if force is False:
        exist, warnings = _check_exist(isamAppliance, cors_policy_name)

    if force is True or exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an access policy",
                "{0}/{1}".format(uri, cors_policy_name),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, check_mode=False, force=False):
    """
    Delete all existing API Access Control Policies
    """
    ret_obj = get_all(isamAppliance)

    if force is True or ret_obj['data'] != []:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete all existing API Access Control Policies",
                "{0}".format(uri),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete_selection(isamAppliance, policies, command="DELETE", check_mode=False, force=False):
    """
    Delete a selection of API Access Control Policies
    """
    found_any = False
    new_list_policies = []

    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']

    for policy in policies:
        found = False
        for obj in ret_obj['data']:
            if obj['name'] == policy:
                found_any = True
                found = True
        if found is False:
            warnings.append("Did not find policy {0} to delete".format(policy))
        else:
            new_list_policies.append(policy)

    if force is True or found_any is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Delete a selection of API Access Control Policies",
                "{0}".format(uri),
                {
                    'command': command,
                    'policies': new_list_policies
                },
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_exist(isamAppliance, cors_policy_name):
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == cors_policy_name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check(isamAppliance, cors_policy_name, allowed_origins, allow_credentials, exposed_headers,
           handle_preflight, allowed_methods, allowed_headers, max_age):
    ret_obj = get_all(isamAppliance)
    exist = False
    json_data = {}

    for obj in ret_obj['data']:
        if obj['name'] == cors_policy_name:
            current_data = obj
            exist = True

    if exist is False:
        return False, ret_obj['warnings'], json_data

    json_data = {
        "name": cors_policy_name,
        "allowed_origins": allowed_origins,
        "allow_credentials": allow_credentials,
        "exposed_headers": exposed_headers,
        "handle_preflight": handle_preflight,
        "allowed_methods": allowed_methods,
        "allowed_headers": allowed_headers,
        "max_age": max_age
    }

    sorted_obj1 = tools.json_sort(json_data)
    logger.debug("Sorted input: {0}".format(sorted_obj1))
    sorted_obj2 = tools.json_sort(current_data)
    logger.debug("Sorted existing data: {0}".format(sorted_obj2))
    if sorted_obj1 != sorted_obj2:
        logger.info("Changes detected, update needed.")
        return True, ret_obj['warnings'], json_data
    else:
        return False, ret_obj['warnings'], json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cors policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/user-info"
requires_modules = None
requires_version = "9.0.3"


# TODO - WARNING: This has not been fully tested.  Please open a bug issue on GITHub if you find a problem.

def get_all(isamAppliance, count=None, start=None, check_mode=False, force=False):
    """
    Retrieve a list of attributes for all users
    """
    return isamAppliance.invoke_get("Retrieve a list of access policies",
                                    "{0}/{1}".format(uri, tools.create_query_string(count=count, start=start)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, userid, check_mode=False, force=False):
    """
    Retrieve the attributes of a user
    """

    return isamAppliance.invoke_get("Retrieve the attributes of a user",
                                    "{0}/{1}".format(uri, userid),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def delete(isamAppliance, userid, check_mode=False, force=False):
    """
    Delete user attributes
    """

    ret_obj = get(isamAppliance, userid)

    if force is True or ret_obj['data'] != []:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete user attributes",
                                               "{0}/{1}".format(uri, userid),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version)

    return isamAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

# URI for this module
uri = "/wga/reverseproxy"
requires_model = "Appliance"

def get_all(isamAppliance, instance_id, check_mode=False, force=False):
    """
    Retrieving all trace components and details - Reverse Proxy
    """
    try:
        return isamAppliance.invoke_get("Retrieving all trace components - Reverse Proxy",
                                        "{0}/{1}/tracing".format(uri, instance_id), requires_model=requires_model)
    except:
        # Return empty array - exception thrown if list has no entries or does not exist
        ret_obj = isamAppliance.create_return_object()
        ret_obj['data'] = {}
        return ret_obj


def get_all_logs(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Retrieving all trace log files for a component - Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieving all trace log files for a component - Reverse Proxy",
                                    "{0}/{1}/tracing/{2}/trace_files".format(uri,
                                                                             instance_id,
                                                                             component_id),requires_model=requires_model)


def get(isamAppliance, instance_id, component_id, file_id, options=None, size=None, start=None, check_mode=False,
        force=False):
    """
    Retrieving snippet of a trace log file for a component - Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieving snippet of a trace log file - Reverse Proxy",
                                    "{0}/{1}/tracing/{2}/trace_files/{3}".format(uri,
                                                                                 instance_id,
                                                                                 component_id,
                                                                                 file_id,
                                                                                 tools.create_query_string(
                                                                                     options=options, start=start,
                                                                                     size=size)),requires_model=requires_model)


def export_file(isamAppliance, instance_id, component_id, file_id, filename, check_mode=False, force=False):
    """
    Exporting a trace log file for a component - Reverse Proxy
    """
    import os.path

    if force is True or (os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file("Exporting a Reverse Proxy trace log file.",
                                                 "{0}/{1}/tracing/{2}/trace_files/{3}?export".format(uri,
                                                                                                     instance_id,
                                                                                                     component_id,
                                                                                                     file_id), filename,requires_model=requires_model)

    return isamAppliance.create_return_object()


def set(isamAppliance, instance_id, component_id, level, flush_interval,
        rollover_size, max_rollover_files, compress, check_mode=False, force=False):
    """
    Modify the trace settings for a component
    """
    warnings = _check(isamAppliance,instance_id)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True,warnings=warnings)
    else:
        return isamAppliance.invoke_put(
            "Modify trace settings for a component",
            "{0}/{1}/tracing/{2}".format(uri,
                                         instance_id,
                                         component_id),
            {
                'level': level,
                'flush_interval': flush_interval,
                'rollover_size': rollover_size,
                'max_rollover_files': max_rollover_files,
                'compress': compress
            },requires_model=requires_model)


def delete(isamAppliance, instance_id, component_id, file_id, check_mode=False, force=False):
    """
    Deleting a trace log file or rollover file for a component - Reverse Proxy
    """
    warnings = _check(isamAppliance,instance_id)

    if force is False:
        try:
            ret_obj = get(isamAppliance, instance_id, component_id, file_id)
            delete_required = True  # Exception thrown if the file is empty
        except:
            delete_required = False

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a trace log file",
                "{0}/{1}/tracing/{2}/trace_files/{3}".format(uri,
                                                             instance_id,
                                                             component_id,
                                                             file_id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Deleting all trace files and rollover files for a component - Reverse Proxy
    """
    warnings = _check(isamAppliance,instance_id)

    if force is False:
        try:
            ret_obj = get_all_logs(isamAppliance, instance_id, component_id)
            delete_required = True  # Exception thrown if the file is empty
        except:
            delete_required = False

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Deleting all trace log files",
                "{0}/{1}/tracing/{2}/trace_files".format(uri,
                                                         instance_id,
                                                         component_id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_multiple_files(isamAppliance, instance_id, component_id, files, check_mode=False, force=False):
    """
    Deleting multiple trace files for a component
    """
    delete_required = False
    files_to_delete = []

    ret_obj = get_all_logs(isamAppliance, instance_id, component_id)
    check_value,warnings = _check(isamAppliance,instance_id)

    if check_value is True:
        for obj1 in files:
            for obj2 in ret_obj['data']:
                if obj1['name'] == obj2['id']:
                    files_to_delete.append(obj1)
                    delete_required = True

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            if len(files_to_delete) == 1:
                return delete(isamAppliance, instance_id, component_id, files_to_delete[0]['name'])
            else:
                return isamAppliance.invoke_put(
                    "Deleting multiple trace files for a component",
                    "{0}/{1}/tracing/{2}/trace_files/?action=delete".format(uri, instance_id, component_id),
                    {
                        'files': files_to_delete
                    }
                ,requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)

def _check(isamAppliance,instance_id):
    """
    Check if it's appliance or not
    :param isamAppliance:
    :return: true|false, warnings message
    """
    ret_obj = get_all(isamAppliance,instance_id)
    check_value, warnings=False, ret_obj['warnings']

    if warnings == []:
        check_value = True
        return check_value, warnings
    else:
        return check_value, warnings



import logging
import copy
import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/isam/cluster/v2"
requires_modules = None
requires_version = None
requires_model = "Docker"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current database configuration
    """
    return isamAppliance.invoke_get("Retrieve the current database configuration", uri,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def set(isamAppliance, hvdb_db_type, hvdb_address, hvdb_port, hvdb_user, hvdb_password, hvdb_db_name=None,
        hvdb_db_secure=None, hvdb_driver_type=None, hvdb_db2_alt_address=None, hvdb_db2_alt_port=None,
        hvdb_solid_tc=None, check_mode=False, force=False, ignore_password_for_idempotency=False):
    """
    Set cluster configuration
    """

    if hvdb_db_type != "soliddb":
        if hvdb_db_name is None:
            return isamAppliance.create_return_object(
                warnings="hvdb_db_type {0} requires hvdb_db_name".format(hvdb_db_type))
        if hvdb_db_secure is None:
            return isamAppliance.create_return_object(
                warnings="hvdb_db_type {0} requires hvdb_db_secure".format(hvdb_db_type))

    if hvdb_db_type == "oracle":
        if hvdb_driver_type is None:
            return isamAppliance.create_return_object(
                warnings="hvdb_db_type {0} requires hvdb_driver_type".format(hvdb_db_type))

    if isinstance(hvdb_port, basestring):
        hvdb_port = int(hvdb_port)

    # Create a simple json with just the main client attributes
    db_json = {
        "hvdb_db_type": hvdb_db_type,
        "hvdb_address": hvdb_address,
        "hvdb_port": hvdb_port,
        "hvdb_user": hvdb_user,
        "hvdb_password": hvdb_password
    }
    # Add attributes that have been supplied... otherwise skip them.

    if hvdb_db2_alt_address is not None:
        db_json["hvdb_db2_alt_address"] = hvdb_db2_alt_address
    if hvdb_db2_alt_port is not None:
        db_json["hvdb_db2_alt_port"] = hvdb_db2_alt_port
    if hvdb_db_name is not None:
        db_json["hvdb_db_name"] = hvdb_db_name
    if hvdb_db_secure is not None:
        db_json["hvdb_db_secure"] = hvdb_db_secure
    if hvdb_driver_type is not None:
        db_json["hvdb_driver_type"] = hvdb_driver_type
    if hvdb_solid_tc is not None:
        if (isinstance(hvdb_solid_tc, basestring)):
            import ast
            hvdb_solid_tc = ast.literal_eval(hvdb_solid_tc)
        db_json["hvdb_solid_tc"] = hvdb_solid_tc

    obj = _check(isamAppliance=isamAppliance, db_json=db_json,
                 ignore_password_for_idempotency=ignore_password_for_idempotency)

    if force is True or obj['value'] is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=obj['warnings'])
        else:
            return isamAppliance.invoke_post("Set database configuration", uri, db_json,
                                             requires_modules=requires_modules, requires_version=requires_version,
                                             requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=obj['warnings'])


def _check(isamAppliance, db_json, ignore_password_for_idempotency=False):
    """
    Check if provided json values match the configuration on appliance

    :param isamAppliance:
    :param db_json:
    :return:
    """

    obj = {'value': True, 'warnings': ""}

    ret_obj = get(isamAppliance)

    obj['warnings'] = ret_obj['warnings']
    del_password = False

    if ignore_password_for_idempotency is True:
        if 'hvdb_password' in ret_obj['data']:
            del ret_obj['data']['hvdb_password']
        if 'hvdb_password' in db_json:
            password = db_json['hvdb_password']
            del_password = True
            del db_json['hvdb_password']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(db_json)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))

    if del_password is True:
        db_json['hvdb_password'] = password

    if sorted_ret_obj != sorted_json_data:
        obj['value'] = False
        return obj
    else:
        obj['value'] = True
        return obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cluster configuration between two appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import os
import shutil
from ibmsecurity.utilities.tools import get_random_temp_dir, files_same
from io import open

logger = logging.getLogger(__name__)
uri = "/isam/runtime_components/pdmgrd/tracing_configuration/v1"
requires_modules = None
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the tracing configuration file for the policy server

    """
    return isamAppliance.invoke_get("Retrieve the tracing configuration file for the policy server",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def export_file(isamAppliance, filepath, check_mode=False, force=False):
    """
    Exporting the tracing configuration file for the policy server
    """

    if os.path.exists(filepath) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filepath))
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_get_file(
            "Exporting the tracing configuration file for the policy server",
            "{0}?export".format(uri), filepath
        )

    return isamAppliance.create_return_object()


def update(isamAppliance, contents, check_mode=False, force=False):
    """
    Update the tracing configuration file for the policy server

    """
    update_required = False

    ret_obj = get(isamAppliance)
    ret_contents = ret_obj['data']['contents']

    if ret_contents.strip() != contents.strip():
        update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_put(
                "Update the tracing configuration file for the policy server",
                "{0}".format(uri),
                {
                    'contents': contents
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def import_file(isamAppliance, filepath, check_mode=False, force=False):
    """
    Update the tracing configuration file of an existing instance

    """

    if force is True or _check_import(isamAppliance=isamAppliance, filepath=filepath) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            with open(filepath, 'r') as infile:
                contents = infile.read()

            return isamAppliance.invoke_put(
                "Updating an authorization server runtime configuration file",
                "{0}".format(uri),
                {
                    'contents': contents.strip()
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def _check_import(isamAppliance, filepath):
    """
    Checks if the file to be imported is the same as the file that's already on the instance
    """
    tmpdir = get_random_temp_dir()
    tmp_original_file = os.path.join(tmpdir, os.path.basename("tempfile.txt"))

    export_file(isamAppliance, filepath=tmp_original_file)

    if files_same(tmp_original_file, filepath):
        logger.debug("files are the same, so we don't want to do anything")
        shutil.rmtree(tmpdir)
        return False
    else:
        logger.debug("files are different, so we return True to indicate the new file should be imported")
        shutil.rmtree(tmpdir)
        return True



import logging
from ibmsecurity.utilities.tools import json_sort
from ibmsecurity.isam.aac.scim.general import get

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"

def set(isamAppliance, schema_name, scim_attribute, mode, scim_subattribute=None, check_mode=False,
                force=False):
    """
    Updating the mode of a SCIM attribute
    """

    ret_obj = get(isamAppliance)

    mode = mode.lower()

    objs = ret_obj['data']['attribute_modes']

    update_required = True

    if scim_subattribute is None:
        obj1 = {'mode': mode, 'attribute': scim_attribute}
    else:
        obj1 = {'mode': mode, 'attribute': scim_attribute, 'subattribute': scim_subattribute}

    obj1 = json_sort(obj1)

    for obj in objs:
        schema = obj['schema']
        if schema == schema_name:
            modes = obj['modes']
            for anitem in modes:
                obj2 = json_sort(anitem)
                if obj1 == obj2:
                    update_required = False

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            if scim_subattribute is None:
                return isamAppliance.invoke_put("Updating the mode of a SCIM attribute",
                                                "{0}/general/attribute_modes/{1}/{2}".format(uri, schema_name,
                                                                                             scim_attribute),
                                                {'mode': mode},
                                                requires_modules=requires_modules,
                                                requires_version=requires_version
                                                )
            else:
                return isamAppliance.invoke_put("Updating the mode of a SCIM attribute",
                                                "{0}/general/attribute_modes/{1}/{2}/{3}".format(uri, schema_name,
                                                                                                 scim_attribute,
                                                                                                 scim_subattribute),
                                                {'mode': mode},
                                                requires_modules=requires_modules,
                                                requires_version=requires_version
                                                )

    return isamAppliance.create_return_object(changed=False)


def reset(isamAppliance, schema_name, scim_attribute, scim_subattribute=None, check_mode=False, force=False):
    """
    Resetting a SCIM attribute mode to default
    """
    ret_obj = get(isamAppliance)

    objs = ret_obj['data']['attribute_modes']

    update_required = False

    for obj in objs:
        schema = obj['schema']
        if schema == schema_name:
            modes = obj['modes']
            for anitem in modes:
                if anitem['attribute'] == scim_attribute:
                    if scim_subattribute is not None:
                        if 'subattribute' in anitem:
                            if anitem['subattribute'] == scim_subattribute:
                                update_required = True
                    elif ('subattribute' in anitem) is False:
                        update_required = True


    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if scim_subattribute is None:
                return isamAppliance.invoke_delete("Resetting a SCIM attribute mode to default",
                                               "{0}/general/attribute_modes/{1}/{2}".format(uri, schema_name,
                                                                                            scim_attribute),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version
                                               )
            else:
                return isamAppliance.invoke_delete("Resetting a SCIM attribute mode to default",
                                               "{0}/general/attribute_modes/{1}/{2}/{3}".format(uri, schema_name,
                                                                                                scim_attribute,
                                                                                                scim_subattribute),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version
                                               )

    return isamAppliance.create_return_object(changed=False)


import logging

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all current LTPA Key files
    """
    return isamAppliance.invoke_get("Retrieving all LTPA Keys",
                                    "/wga/ltpa_key")


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a LTPA Key file
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a LTPA key file",
                "/wga/ltpa_key/{0}".format(id))

    return isamAppliance.create_return_object()


def rename(isamAppliance, id, new_name, check_mode=False, force=False):
    """
    Rename a LTPA key file
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Rename a LTPA Key file",
                "/wga/ltpa_key/{0}".format(id),
                {
                    'id': id,
                    'new_name': new_name
                })

    return isamAppliance.create_return_object()


def export_key(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Exporting a LTPA Key
    """
    import os.path

    if force is True or (_check(isamAppliance, id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a LTPA key file",
                "/wga/ltpa_key/{0}?export".format(id),
                filename)

    return isamAppliance.create_return_object()


def import_key(isamAppliance, id, ltpa_keyfile, check_mode=False, force=False):
    """
    Importing a LTPA key file
    """
    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a LTPA key file",
                "/wga/ltpa_key/",
                [
                    {
                        'file_formfield': 'ltpa_keyfile',
                        'filename': ltpa_keyfile,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {})

    return isamAppliance.create_return_object()


def _check(isamAppliance, id):
    """
    Check if URL Mapping already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['id'] == id:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare URL Mapping between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['version']
    for obj in ret_obj2['data']:
        del obj['version']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging
import os.path
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve existing fixpacks.
    """
    return isvgAppliance.invoke_get("Retrieving fixpacks",
                                    "/fixpacks")


def install(isvgAppliance, file, check_mode=False, force=False):
    """
    Install fixpack
    """
    if force is True or _check(isvgAppliance, file) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post_files(
                "Install fixpack",
                "/fixpacks",
                [{
                    'file_formfield': 'file',
                    'filename': file,
                    'mimetype': 'application/octet-stream'
                }],
                {})

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, fixpack):
    """
    Check if fixpack is already installed
    """
    ret_obj = get(isvgAppliance)

    fixpack_name = _extract_fixpack_name(fixpack)

    # Reverse sort the json by 'id'
    json_data_sorted = sorted(ret_obj['data'], key=lambda k: int(k['id']), reverse=True)
    # Eliminate all rollbacks
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fixpack in json_data_sorted:
        if fixpack['action'] == 'Uninstalled':
            del_fixpack = fixpack['name']
        elif del_fixpack == fixpack['name'] and fixpack['rollback'] == 'Yes':
            del_fixpack = ''
        elif fixpack['name'].lower() == fixpack_name.lower():
            return True

    return False


def _extract_fixpack_name(fixpack):
    """
    Extract fixpack name from the given fixpack
    """
    import re

    # Look for the follwing string inside the fixpack file
    # FIXPACK_NAME="9021_IPv6_Routes_fix"
    for s in ibmsecurity.utilities.tools.strings(fixpack):
        match_obj = re.search(r"FIXPACK_NAME=\"(?P<fp_name>\w+)\"", s)
        if match_obj:
            logger.info("Fixpack name extracted from file using strings method: {0}".format(match_obj.group('fp_name')))
            return match_obj.group('fp_name')

    # Unable to extract fixpack name from binary
    # Return fixpack name derived from the filename
    logger.debug(fixpack)
    file_name = os.path.basename(fixpack)
    fixpack_name, ext_name = file_name.split('.fixpack')
    logger.debug("Extracted fixpack_name: " + fixpack_name)
    return fixpack_name


def compare(isvgAppliance1, isvgAppliance2):
    """
    Compare fixpacks between two appliances
    Sort in reverse order and remove fixpacks that were rolled back before compare
    """
    ret_obj1 = get(isvgAppliance1)
    ret_obj2 = get(isvgAppliance2)

    json_data1_sorted = sorted(ret_obj1['data'], key=lambda k: int(k['id']), reverse=True)
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fixpack in json_data1_sorted:
        if fixpack['action'] == 'Uninstalled':
            del_fixpack = fixpack['name']
            del fixpack
        elif del_fixpack == fixpack['name'] and fixpack['rollback'] == 'Yes':
            del_fixpack = ''
            del fixpack
        else:
            del fixpack['date']

    json_data2_sorted = sorted(ret_obj2['data'], key=lambda k: int(k['id']), reverse=True)
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fixpack in json_data2_sorted:
        if fixpack['action'] == 'Uninstalled':
            del_fixpack = fixpack['name']
            del fixpack
        elif del_fixpack == fixpack['name'] and fixpack['rollback'] == 'Yes':
            del_fixpack = ''
            del fixpack
        else:
            del fixpack['date']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

uri = "/mga/user_registry/groups"
requires_modules = ["mga", "federation"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all current groups in the registry
    """
    return isamAppliance.invoke_get("Retrieving a list of all current groups in the registry",
                                    "{0}/v1".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieving details for a particular group in the registry
    """
    return isamAppliance.invoke_get("Retrieving details for a particular group in the registry",
                                    "{0}/{1}/v1".format(uri, id),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add_user(isamAppliance, user_name, id, check_mode=False, force=False):
    """
    Adding a user to a group in the registry
    """
    if force is True or _check(isamAppliance, user_name, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Adding a user to a group in the registry",
                "/mga/user_registry/users/{0}/groups/v1".format(user_name),
                {
                    'id': id
                },
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete_user(isamAppliance, user_name, id, check_mode=False, force=False):
    """
    Removing a user from a group in the registry
    """
    if force is True or _check(isamAppliance, user_name, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Removing a user from a group in the registry",
                "/mga/user_registry/users/{0}/groups/{1}/v1".format(user_name, id),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, user_name, id):
    """
    Check if user already exists
    """
    ret_obj = get_all(isamAppliance)

    for grp in ret_obj['data']:
        if grp['id'] == id:
            ret_obj = get(isamAppliance, id)
            for usr in ret_obj['data']['users']:
                if usr['id'] == user_name:
                    return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare groups between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        user = get(isamAppliance1, obj['id'])
        obj['users'] = user['data']['users']
    for obj in ret_obj2['data']:
        user = get(isamAppliance2, obj['id'])
        obj['users'] = user['data']['users']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[''])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

uri = "/wga/reverseproxy"
requires_modules = ["wga"]
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve all reverse proxies
    """
    return isamAppliance.invoke_get(description="Retrieving all reverse proxies", uri=uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def _check(isamAppliance, id):
    """
    Check if reverse proxy is already created or not

    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance)

    logger.debug("Looking for existing reverse proxies in: {0}".format(ret_obj['data']))
    if ret_obj['data']:
        for rp in ret_obj['data']:
            if rp['id'] == id:
                logger.debug("Found reverse proxy: {0}".format(id))
                return True

    return False


def add(isamAppliance, inst_name, admin_pwd, host='localhost', listening_port='7234', domain='Default',
        admin_id='sec_master', ssl_yn='no', key_file=None, cert_label=None, ssl_port=None, http_yn='no', http_port='80',
        https_yn='yes', https_port='443', nw_interface_yn='no', ip_address='0.0.0.0', check_mode=False, force=False):
    """
    Add a reverse proxy

    :param isamAppliance:
    :return:
    """
    if force is True or _check(isamAppliance, inst_name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(description="Add a reverse proxy", uri=uri,
                                             data={
                                                 "inst_name": inst_name,
                                                 "host": host,
                                                 "listening_port": listening_port,
                                                 "domain": domain,
                                                 "admin_id": admin_id,
                                                 "admin_pwd": admin_pwd,
                                                 "ssl_yn": ssl_yn,
                                                 "key_file": key_file,
                                                 "cert_label": cert_label,
                                                 "ssl_port": ssl_port,
                                                 "http_yn": http_yn,
                                                 "http_port": http_port,
                                                 "https_yn": https_yn,
                                                 "https_port": https_port,
                                                 "nw_interface_yn": nw_interface_yn,
                                                 "ip_address": ip_address
                                             },
                                             requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, admin_pwd, admin_id='sec_master', domain='Default', check_mode=False, force=False):
    """
    Unconfigure existing runtime component
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(description="Remove a reverse proxy",
                                            uri="{0}/{1}".format(uri, id),
                                            data={
                                                "operation": "unconfigure",
                                                "admin_id": admin_id,
                                                "admin_pwd": admin_pwd,
                                                "domain": domain
                                            },
                                            requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def import_config(isamAppliance, id, file, overwrite=True, check_mode=False, force=False):
    """
    Import or migrate reverse proxy
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(description="Import or Migrate reverse proxy",
                                                   uri="{0}/{1}/migrate".format(uri, id),
                                                   fileinfo=[{
                                                       'file_formfield': 'file',
                                                       'filename': file,
                                                       'mimetype': 'application/octet-stream'
                                                   }],
                                                   data={'overwrite': overwrite},
                                                   requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def export_config(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Export a Reverse Proxy Configuration
        -id is the name of the reverse proxy instance
        -filename is file system location to export the file (e.g. /tmp/webseal_config.zip)
    """
    import os.path
    if force is True or os.path.exists(os.path.dirname(filename)) is True:
        if check_mode is True:  # No point downloading a file if in check_mode
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_get_file(
                description="Export a Reverse Proxy Configuration",
                uri="{0}/{1}?action=export".format(uri, id),
                filename=filename)

    return isamAppliance.create_return_object()


def execute(isamAppliance, id, operation="restart", check_mode=False, force=False):
    """
    Execute an operation on runtime component

    :param isamAppliance:
    :param operation:
    :return:
    """
    ret_obj = get(isamAppliance)

    for rp in ret_obj['data']:
        if rp['id'] == id:
            if (force is True or
                    (rp['restart'] == "true" and operation == "restart") or
                    (rp['started'] == 'yes' and operation == "stop") or
                    (rp['started'] == 'no' and operation == "start")):
                if check_mode is True:
                    return isamAppliance.create_return_object(changed=True)
                else:
                    return isamAppliance.invoke_put(description="Execute an operation on reverse proxy",
                                                    uri="{0}/{1}".format(uri, id),
                                                    data={
                                                        "operation": operation
                                                    },
                                                    requires_modules=requires_modules,
                                                    requires_version=requires_version)
            break

    return isamAppliance.create_return_object()


def execute_multiples(isamAppliance, instances, operation, check_mode=False, force=False):
    """
    Stopping, starting, or restarting multiple instances

    :param isamAppliance:
    :param operation:
    :return:
    """
    ret_obj = get(isamAppliance)
    new_instances = []

    for instance in instances:
        for rp in ret_obj['data']:
            if rp['id'] == instance['instance_name']:
                if ((rp['restart'] == "true" and operation == "restart") or
                        (rp['started'] == 'yes' and operation == "stop") or
                        (rp['started'] == 'no' and operation == "start")):
                    new_instances.append(instance)

    if force is True or len(new_instances) >= 1:

        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        elif len(new_instances) > 1:
            return isamAppliance.invoke_put("Stopping, starting, or restarting multiple instances",
                                            "{0}".format(uri),
                                            {
                                                "operation": operation,
                                                "instances": new_instances
                                            },
                                            requires_modules=requires_modules,
                                            requires_version=requires_version)
        elif len(new_instances) == 1:
            execute(isamAppliance, new_instances[0]['instance_name'], operation=operation, check_mode=check_mode,
                    force=force)

    return isamAppliance.create_return_object()


def obfuscating(isamAppliance, id, pwd, check_mode=False, force=False):
    """
    Obfuscating a GSO password
    Must set gso-obfuscation-key in the config file for the password to be obfuscated.
    https://www.ibm.com/support/knowledgecenter/SSPREK_9.0.6/com.ibm.isam.doc/wrp_stza_ref/reference/ref_gso_obfuscation_key.html
    """
    return isamAppliance.invoke_post("Obfuscating a GSO password",
                                     "{0}/{1}?action=obfuscate_gso_pwd".format(uri, id),
                                     {
                                         "pwd": pwd
                                     },
                                     requires_modules=requires_modules, requires_version=requires_version)


def renew_cert(isamAppliance, id, isamUser, check_mode=False, force=False):
    """
    Renew a reverse proxy instance management certificate
    """
    return isamAppliance.invoke_put("Renew a reverse proxy instance management certificate",
                                    "{0}/{1}".format(uri, id),
                                    {
                                        "admin_id": isamUser.username,
                                        "admin_pwd": isamUser.password,
                                        "operation": "renew"
                                    },
                                    requires_modules=requires_modules, requires_version=requires_version)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare list of reverse proxies between 2 appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for rp in ret_obj1['data']:
        del rp['started']
        del rp['restart']
        del rp['version']
    for rp in ret_obj2['data']:
        del rp['started']
        del rp['restart']
        del rp['version']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['restart', 'started', 'version'])



from setuptools import setup, find_packages

setup(
    name = 'ibmsecurity',
    packages = find_packages(),
    # Date of release used for version - please be sure to use YYYY.MM.DD.seq#, MM and DD should be two digits e.g. 2017.02.05.0
    # seq# will be zero unless there are multiple release on a given day - then increment by one for additional release for that date
    version = '2023.4.25.0',
    description = 'Idempotent functions for IBM Security Appliance REST APIs',
    author='IBM',
    author_email='secorch@us.ibm.com',
    url='',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Topic :: Software Development :: Build Tools'
    ],
    zip_safe = False,
    install_requires = [
        'requests',
        'jmespath'
    ]
)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/devicefingerprints/userIds/"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of user IDs from devices fingerprints
    """
    return isamAppliance.invoke_get("Retrieve a list of user IDs from Devices Fingerprints",
                                    uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def search(isamAppliance, userID, force=False, check_mode=False):
    """
    Search device id by userId
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['userID'] == userID:
            logger.info("Found user {0}".format(userID))
            return_obj['data'] = userID
            return_obj['rc'] = 0

    return return_obj


def get(isamAppliance, userID, check_mode=False, force=False):
    """
    Retrieve a list of device fingerprints for a given user ID
    """
    warnings = []
    ret_obj = search(isamAppliance, userID=userID, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("User {0} had no match, skipping retrieval.".format(userID))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        return isamAppliance.invoke_get("Retrieve a list of device fingerprints for a given user ID",
                                        "{0}/{1}".format(uri, userID))


def delete(isamAppliance, userID, check_mode=False, force=False):
    """
    Delete all device fingerprints for the given user ID
    """
    warnings = []
    ret_obj = search(isamAppliance, userID=userID, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("User {0} had no match, skipping Delete.".format(userID))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete all device fingerprints for the given user ID",
                                               "{0}/{1}".format(uri, userID))



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/fido2/metadata"
requires_modules = ["mga"]
requires_version = "9.0.7.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of FIDO2 Metadata
    """
    return isamAppliance.invoke_get("Retrieving the list of FIDO2 Metadata", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)

def get(isamAppliance, name, id=None, check_mode=False, force=False):
    """
    Retrieve a specific FIDO2 Metadata
    """
    if id is None:
        ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
        id = ret_obj['data']

    if id == {}:
        logger.info("FIDO2 Metadata {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)

def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search FIDO2 Metadata id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['filename'] == name:
            logger.info("Found FIDO2 Metadata {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj

def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a FIDO2 Metadata file
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("FIDO2 Metadata {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a FIDO2 Metadata file",
                "{0}/{1}".format(uri, id),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def export_file(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Export a specific FIDO2 Metadata file
    """
    import os.path
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("FIDO2 Metadata {0} not found, skipping export.".format(name))
    else:
        if force is True or (os.path.exists(filename) is False):
            if check_mode is False:  # No point downloading a file if in check_mode
                return isamAppliance.invoke_get_file(
                    "Export a specific FIDO2 Metadata file",
                    "{0}/{1}/file".format(uri,id),
                    filename,
                    requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def import_file(isamAppliance, filename, check_mode=False, force=False):
    """
    Import a new FIDO2 Metadata file or Import FIDO2 Metadata file contents to existing

    Only a valid FIDO MDS Document (extension .json), Yubico Metadata (extension .yubico), or PEM Certificate (extension .pem) will be accepted.
    Any other file type will fail validation.
    """
    import_new = False
    update_required = False
    if force is False:
        ret_obj = search(isamAppliance, _extract_filename(filename))
        if ret_obj['data'] != {}:
            ret_obj_content = _get(isamAppliance, ret_obj['data'])
            with open(filename, 'r') as infile:
                content = infile.read()
            if (ret_obj_content['data']['contents']).strip() != content.strip():
                update_required = True
        else:
            import_new = True

    if force is True or update_required is True or import_new is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if import_new is True:
                filebasename = _extract_filename(filename)
                return isamAppliance.invoke_post_files(
                "Import a new FIDO2 Metadata file",
                "{0}".format(uri),
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/file'
                    }
                ],
                {
                    "name": filebasename,
                    "filename": filebasename
                },
                requires_modules=requires_modules, requires_version=requires_version)
            else:
                return isamAppliance.invoke_post_files(
                    "Import a FIDO2 Metadata file (replace)",
                    "{0}/{1}/file".format(uri,ret_obj['data']),
                    [
                        {
                            'file_formfield': 'file',
                            'filename': filename,
                            'mimetype': 'application/file'
                        }
                    ],
                    {},
                    requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific FIDO2 Metadata by id",
                                    f"{uri}/{id}",
                                    requires_modules=requires_modules, requires_version=requires_version)

def _extract_filename(upload_filename):
    """
    Extract filename from fully qualified path to use if no filename provided
    """
    import os.path
    return os.path.basename(upload_filename)

def _check(isamAppliance, fido2_metadata):
    """
    Check if FIDO2 metadata configuration is identical with server
    """
    ret_obj = get_all(isamAppliance)
    logger.debug("Comparing server FIDO2 metadata configuration with desired configuration.")
    # Converting python ret_obj['data'] dict to valid JSON (RFC 8259)
    # e.g. converts python boolean 'True' -> to JSON literal lowercase value 'true'
    cur_json_string = json.dumps(ret_obj['data']) 
    cur_sorted_json = tools.json_sort(cur_json_string)
    logger.debug("Server JSON : {0}".format(cur_sorted_json))
    given_json_string = json.dumps(fido2_metadata)
    given_sorted_json = tools.json_sort(given_json_string)
    logger.debug("Desired JSON: {0}".format(given_sorted_json))
    if cur_sorted_json != given_sorted_json:
        return False
        logger.debug("Changes detected!")
    else:
        logger.debug("Server configuration is identical with desired configuration. No change necessary.")
        return True



import logging
import ibmsecurity.utilities.tools
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)
requires_model="Appliance"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all static routes
    """
    return isamAppliance.invoke_get("Retrieving all static routes",
                                    "/net/routes", requires_model=requires_model)


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Retrieving a single static route
    """
    return isamAppliance.invoke_get("Retrieving a single static route",
                                    "/net/routes/{0}".format(uuid), requires_model=requires_model)


def add(isamAppliance, address, enabled=True, comment=None, table='main', maskOrPrefix=None, gateway=None, label=None,
        vlanId=None, metric=None, check_mode=False, force=False):
    """
    Creating a static route
    """

    if isamAppliance.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance.facts['model'])]
        return isamAppliance.create_return_object(warnings=warnings)

    if table.lower() != 'main':
        table_uuid = ibmsecurity.isam.base.network.interfaces.search(isamAppliance, address=table)
        if table_uuid['data'] != {}:
            table = table_uuid['data']
        else:
            ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
            found_table = False
            interfaces = ret_obj['data']['interfaces']
            for obj in interfaces:
                items = obj['ipv4']['addresses']
                for item in items:
                    if item['uuid'] == table:
                        found_table = True
            if found_table is False:
                logger.debug("Route table {0} is not found, Add static route is not supported.".format(table))
                return isamAppliance.create_return_object(changed=False)

    interfaceUUID = _get_interfaceUUID(isamAppliance, label, vlanId)
    if interfaceUUID is None and label is not None:
        logger.debug("Interface {0} not found, Add static route is not supported.".format(label))
        return isamAppliance.create_return_object(changed=False)

    if maskOrPrefix is not None:
        if isinstance(maskOrPrefix, basestring):
            if maskOrPrefix.lower() == 'none':
                maskOrPrefix = ''
    if isinstance(table, basestring):
        if table.lower() == 'none':
            table = None
    if isinstance(metric, basestring):
        if metric.lower() == 'none':
            metric = None
        else:
            metric = int(metric)

    if force is True or _check(isamAppliance=isamAppliance, address=address, table=table,
                               interfaceUUID=interfaceUUID) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a static route",
                "/net/routes",
                {
                    "_isNew": True,
                    "enabled": enabled,
                    "address": address,
                    "maskOrPrefix": maskOrPrefix,
                    "gateway": gateway,
                    "interfaceUUID": interfaceUUID,
                    "metric": metric,
                    "comment": comment,
                    "table": table
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, address, new_address=None, enabled=True, maskOrPrefix=None, gateway=None,
           metric=None, comment=None, table='main', label=None, vlanId=None, new_label=None, new_vlanId=None,
           check_mode=False, force=False):
    """
    Updating a static route
    """
    if isamAppliance.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance.facts['model'])]
        return isamAppliance.create_return_object(warnings=warnings)

    warnings = []
    interfaceUUID = _get_interfaceUUID(isamAppliance, label, vlanId)
    if interfaceUUID is None and label is not None:
        warnings.append("Interface {0} not found, Update static route is not supported.".format(label))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if table.lower() != 'main':
        table_uuid = ibmsecurity.isam.base.network.interfaces.search(isamAppliance, address=table)
        if table_uuid['data'] != {}:
            table = table_uuid['data']
        else:
            ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
            found_table = False
            interfaces = ret_obj['data']['interfaces']
            for obj in interfaces:
                items = obj['ipv4']['addresses']
                for item in items:
                    if item['uuid'] == table:
                        found_table = True
            if found_table is False:
                logger.debug("Route table {0} is not found, Add static route is not supported.".format(table))
                return isamAppliance.create_return_object(changed=False)

    uuid = _get_uuid(isamAppliance, address, table, interfaceUUID)

    if uuid is None:
        logger.info("Unable to find Static Route to modify: {0} / {1}".format(address, table))
        return isamAppliance.create_return_object()

    if new_label is not None:
        interfaceUUID = _get_interfaceUUID(isamAppliance, new_label, new_vlanId)
    if interfaceUUID is None:
        interfaceUUID = ''
    if new_address is not None:
        address = new_address
    if maskOrPrefix is not None:
        if isinstance(maskOrPrefix, basestring):
            if maskOrPrefix.lower() == 'none':
                maskOrPrefix = ''

    if isinstance(metric, basestring):
        if metric.lower() == 'none':
            metric = None
        else:
            metric = int(metric)
    if isinstance(enabled, basestring):
        if enabled.lower() == 'true':
            enabled = True
        else:
            enabled = False

    json_data = {
        "enabled": enabled,
        "address": address,
        "maskOrPrefix": maskOrPrefix,
        "gateway": gateway,
        "interfaceUUID": interfaceUUID,
        "metric": metric,
        "comment": comment,
        "table": table,
        "objType": "staticRoute",
        "uuid": uuid
    }

    if force is True or _compare(isamAppliance, json_data) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating a static route",
                "/net/routes/{0}".format(uuid), json_data)

    return isamAppliance.create_return_object()


def set(isamAppliance, address, new_address=None, enabled=True, maskOrPrefix=None, gateway=None, metric=None,
        comment=None, table='main', label=None, vlanId=None, new_label=None, new_vlanId=None, check_mode=False,
        force=False):

    if isamAppliance.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance.facts['model'])]
        return isamAppliance.create_return_object(warnings=warnings)

    if table != 'main':
        table_uuid = ibmsecurity.isam.base.network.interfaces.search(isamAppliance, address=table)
        table_uuid = table_uuid['data']
    else:
        table_uuid = 'main'

    warnings = []
    interfaceUUID = _get_interfaceUUID(isamAppliance, label, vlanId)
    if interfaceUUID is None and label is not None:
        warnings.append("Interface {0} not found, Update static route is not supported.".format(label))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if _check(isamAppliance, address, table_uuid, interfaceUUID) is True:
        return update(isamAppliance=isamAppliance, address=address, new_address=new_address, enabled=enabled,
                      maskOrPrefix=maskOrPrefix, gateway=gateway, label=label,
                      vlanId=vlanId, new_label=new_label, new_vlanId=new_vlanId, metric=metric, comment=comment,
                      table=table, check_mode=check_mode, force=force)
    else:
        return add(isamAppliance=isamAppliance, address=address, enabled=enabled, maskOrPrefix=maskOrPrefix,
                   gateway=gateway, label=label, vlanId=vlanId,
                   metric=metric, comment=comment, table=table, check_mode=check_mode, force=force)


def _get_interfaceUUID(isamAppliance, label, vlanId=None):
    if label is None or label == '' or label.lower() == 'auto':
        interfaceUUID = ''
    else:
        intf, warnings= ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if intf:
            interfaceUUID = intf['uuid']
        else:
            interfaceUUID = ''

    return interfaceUUID


def _get_uuid(isamAppliance, address, table, interfaceUUID):
    ret_obj = get_all(isamAppliance)
    for sr in ret_obj['data']['staticRoutes']:
        logger.debug("Scanning {0}/{1}/{2} in Static Routes: {3}".format(address, table, interfaceUUID, sr))
        if sr['address'] == address and (
                sr['interfaceUUID'] == interfaceUUID or (sr['interfaceUUID'] == '' and interfaceUUID is None)) and sr[
            'table'] == table:
            return sr['uuid']

    return None


def delete(isamAppliance, address, table='main', label=None, vlanId=None, check_mode=False, force=False):
    """
    Delete a static route
    """
    if isamAppliance.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance.facts['model'])]
        return isamAppliance.create_return_object(warnings=warnings)

    warnings = []
    interfaceUUID = _get_interfaceUUID(isamAppliance, label, vlanId)

    if interfaceUUID is None and label is not None:
        warnings.append("Interface {0} not found, Delete static route is not supported.".format(label))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if table != 'main':
        table_uuid = ibmsecurity.isam.base.network.interfaces.search(isamAppliance, address=table)
        if table_uuid['data'] != {}:
            table = table_uuid['data']
        else:
            ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
            found_table = False
            interfaces = ret_obj['data']['interfaces']
            for obj in interfaces:
                items = obj['ipv4']['addresses']
                for item in items:
                    if item['uuid'] == table:
                        found_table = True
            if found_table is False:
                logger.debug("Route table {0} is not found, Add static route is not supported.".format(label))
                return isamAppliance.create_return_object(changed=False)

    uuid = _get_uuid(isamAppliance=isamAppliance, address=address, table=table, interfaceUUID=interfaceUUID)

    if force is True or uuid is not None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a static route",
                "/net/routes/{0}".format(uuid))

    return isamAppliance.create_return_object()


def _check(isamAppliance, address, table, interfaceUUID):
    """
    Check if static route already exists
    """
    ret_obj = get_all(isamAppliance)

    for sr in ret_obj['data']['staticRoutes']:
        if sr['address'] == address and sr['table'] == table and sr['interfaceUUID'] == interfaceUUID:
            return True

    return False


def _compare(isamAppliance, json_data):
    """
    Check if static route already exists
    """
    ret_obj = get_all(isamAppliance)

    for sr in ret_obj['data']['staticRoutes']:
        if sr['address'] == json_data['address'] and sr['table'] == json_data['table'] and sr['interfaceUUID'] == \
                json_data['interfaceUUID']:
            logger.debug(ibmsecurity.utilities.tools.json_sort(sr))
            logger.debug(ibmsecurity.utilities.tools.json_sort(json_data))
            return ibmsecurity.utilities.tools.json_sort(sr) == ibmsecurity.utilities.tools.json_sort(json_data)

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare static routes between 2 appliances
    """
    if isamAppliance1.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance1.facts['model'])]
        return isamAppliance1.create_return_object(warnings=warnings)

    if isamAppliance2.facts['model'] != requires_model:
        warnings = ["API invoked requires model: {0}, appliance is of deployment model: {1}.".format(requires_model, isamAppliance2.facts['model'])]
        return isamAppliance2.create_return_object(warnings=warnings)

    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for sr in ret_obj1['data']['staticRoutes']:
        del sr['uuid']
        del sr['interfaceUUID']
    for sr in ret_obj2['data']['staticRoutes']:
        del sr['uuid']
        del sr['interfaceUUID']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid', 'interfaceUUID'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve current version.
    """
    return isamAppliance.invoke_get("Retrieving version",
                                    "/core/sys/versions")



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/configuration/services/"
requires_modules = None
requires_versions = None
requires_model = "Appliance"


def add(isamAppliance, service_name, name, value, check_mode=False, force=False):
    """
    Creates a service attribute
    """

    check_value, warnings = _check(isamAppliance, service_name, name)
    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Creating a service attribute",
                                             "{0}{1}/attributes".format(module_uri, service_name),
                                             {
                                                 "name": name,
                                                 "value": value
                                             }, requires_version=requires_versions,
                                             requires_modules=requires_modules, requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, service_name, attribute_name, check_mode=False, force=False):
    """
    deletes a service level attribute
    """

    check_value, warnings = _check(isamAppliance, service_name, attribute_name)
    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Deleting a service attribute",
                                               "{0}{1}/attributes/{2}".format(module_uri, service_name,
                                                                              attribute_name),
                                               requires_version=requires_versions,
                                               requires_modules=requires_modules, requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def get(isamAppliance, service_name, attribute_name):
    """
    Retrieving a service attribute
    """

    return isamAppliance.invoke_get("Retrieving a service attribute",
                                    "{0}{1}/attributes/{2}".format(module_uri, service_name, attribute_name),
                                    requires_version=requires_versions,
                                    requires_modules=requires_modules,
                                    requires_model=requires_model)


def get_all(isamAppliance, service_name):
    """
    Retrieving service attribute names
    """
    return isamAppliance.invoke_get("Retrieving service attribute names",
                                    "{0}{1}/attributes?includeAllValues=true".format(module_uri, service_name),
                                    requires_version=requires_versions,
                                    requires_modules=requires_modules,
                                    requires_model=requires_model)


def update(isamAppliance, service_name, attribute_name, attribute_value, check_mode=False, force=False):
    """
    Updating a service attribute
    """

    check_value, warnings = _check_add(isamAppliance, service_name, attribute_name, attribute_value)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating a service attribute",
                                            "{0}{1}/attributes/{2}".format(module_uri, service_name, attribute_name),
                                            {
                                                "value": attribute_value
                                            },
                                            requires_modules=requires_modules, requires_version=requires_versions,
                                            requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, service_name, attribute_name, attribute_value, check_mode=False, force=False):
    """
    Determines if add or update is called
    """

    check_value, warnings = _check(isamAppliance, service_name, attribute_name)

    if check_value is False:
        return add(isamAppliance, service_name, attribute_name, attribute_value, check_mode, force)
    else:
        return update(isamAppliance, service_name, attribute_name, attribute_value, check_mode, force)


def compare(isamAppliance1, service_name1, isamAppliance2, service_name2):
    """
    Compare configuration between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, service_name1)
    ret_obj2 = get_all(isamAppliance2, service_name2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])


def _check_add(isamAppliance, service_name, name, value):
    """
    idempotency test for add function
    """
    check_obj = {}
    warnings = ""
    # check to see if attribute under service name exist, return True if it doesnt exist
    try:
        check_obj = get(isamAppliance, service_name, name)
        warnings = check_obj['warnings']
    except:
        return True, warnings

    if 'value' in check_obj['data']:
        if check_obj['data']['value'] != value:
            return True, warnings
        else:
            return False, warnings
    else:
        return False, warnings


def _check(isamAppliance, service_name, attribute_name):
    """
    Checks to see if attribute exists
    """
    warnings = ""
    try:
        check_obj = get(isamAppliance, service_name, attribute_name)
        warnings = check_obj['warnings']
    except:
        return False, warnings

    if check_obj['data'] == {}:
        return False, warnings

    return True, warnings



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/attribute_sources"
requires_modules = ["federation"]
requires_version = "9.0.0.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of attribute sources
    """
    return isamAppliance.invoke_get("Retrieve a list of attribute sources", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific attribute source
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("Attribute Source {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    """
    Internal function to get data using "id" - used to avoid extra calls

    :param isamAppliance:
    :param id:
    :return:
    """
    return isamAppliance.invoke_get("Retrieve a specific attribute source",
                                    "{}/{}".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search Attribute Source by name
    """
    return_obj = isamAppliance.create_return_object()
    ret_obj = get_all(isamAppliance)
    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Attribute Source {0} - id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']

    return return_obj


def set(isamAppliance, name, type, value, properties=None, new_name=None, check_mode=False, force=False):
    """
    Creating or Modifying an Attribute Source
    """
    ret_obj = search(isamAppliance, name)
    as_id = ret_obj['data']

    if as_id == {}:
        # Force the add - we already know attribute source does not exist
        return add(isamAppliance, name, type, value, properties, check_mode, True)
    else:
        # Update request
        return update(isamAppliance, name, type, value, properties, new_name, check_mode, force)


def add(isamAppliance, name, type, value, properties=None, check_mode=False, force=False):
    """
    Create an attribute source
    """
    ret_obj = search(isamAppliance, name, check_mode, force)
    as_id = ret_obj['data']
    if as_id != {}:
        logger.info("Attribute Source {0} already exists.".format(name))

    if force is True or as_id == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "type": type,
                "value": value
            }
            if properties is not None:
                json_data['properties'] = properties
            return isamAppliance.invoke_post(
                "Create an attribute source",
                uri, json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, type, value, properties=None, new_name=None, check_mode=False, force=False):
    """
    Update an attribute source
    """
    as_id, update_required, json_data = _check(isamAppliance, name, type, value, properties, new_name)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update an attribute source",
                "{0}/{1}".format(uri, as_id), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, type, value, properties, new_name=None):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {
        "type": type,
        "value": value
    }
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("Attribute Source not found, returning no update required.")
        return None, update_required, json_data
    else:
        as_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        if properties is not None:
            json_data['properties'] = properties
        else:
            # May not exist so skip any exceptions when deleting
            try:
                del ret_obj['data']['properties']
            except:
                pass
        del ret_obj['data']['id']
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return as_id, update_required, json_data


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an attribute source
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    as_id = ret_obj['data']
    if as_id == {}:
        logger.info("Attribute Source: {0}, no longer exists. Skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an attribute source",
                "{0}/{1}".format(uri, as_id),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Attribute Sources between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/password_vault/"
requires_modules = ["federation", "mga"]
requires_version = "10.0.1"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the password vault configuration
    """
    return isamAppliance.invoke_get("Retrieving the password vault configuration",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def update(isamAppliance, enabled, data_location, resources, admin_group, public_key='', check_mode=False,
           force=False):
    """
    Update a specified Risk Profile
    """

    json_data = {
        'admin_group': admin_group,
        'data_location': data_location,
        'enabled': enabled,
        'resources': resources,
        'public_key': public_key
    }
    update_required, warnings = _check(isamAppliance, json_data)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified Risk Profile",
                "{0}".format(uri), json_data, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, json_data):
    ret_obj = get(isamAppliance)
    sorted_json1 = tools.json_sort(ret_obj['data'])
    sorted_json2 = tools.json_sort(json_data)

    if sorted_json1 == sorted_json2:
        return False, ret_obj['warnings']
    else:
        return True, ret_obj['warnings']



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/partner_templates"
requires_modules = ["federation"]
requires_version = "9.0.0.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Get the status of the federation partner templates
    """
    return isamAppliance.invoke_get("Get the status of the federation partner templates",
                                    uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def upload(isamAppliance, filename, check_mode=False, force=False):
    """
    Update the federation partner templates

    Technote: http://www-01.ibm.com/support/docview.wss?uid=swg21970379
    Description: Connectors available for SAML 2.0

    :param isamAppliance:
    :param filename:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isamAppliance, filename) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Update the federation partner templates", uri,
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {},
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, filename):
    ret_obj = get(isamAppliance)

    cur_ver = ret_obj['data']['version']
    logger.info("Current Installed Version: {0}".format(cur_ver))

    import os
    fname = os.path.splitext(filename)[0]

    new_ver = fname.split('_')[-1]

    logger.info("Provided Package Version: {0}".format(new_ver))

    if new_ver > cur_ver:
        logger.info("Install of new package can proceed")
        return True
    else:
        logger.info("current package is upto date, no need for upgrade at this time.")
        return False



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/certstore_object"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Get personal cert store entries
    """
    return isvgAppliance.invoke_get("Retrieving cert store entries",
                                    uri + "/key/certificates?type=personal")


def search(isvgAppliance, label, check_mode=False, force=False):
    """
    Search for existing certificate by label
    """
    ret_obj = get_all(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if 'id' in obj and obj['label'] == label:
            logger.info("Found cert entry: {0}".format(obj['label']))
            return_obj['data'] = obj
            return_obj['rc'] = 0
            break

    return return_obj
    

def upload(isvgAppliance, db, type, password, label="lmi", check_mode=False, force=False):
    """
    Import personal certificate from db
    """
    warnings = []

    file_short = tools.path_leaf(db);
    if force is True or not _check(isvgAppliance, label):
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_post_files(
                "Import personal certificate",
                uri + "/key/certificates/upload",
                [
                    {
                        'file_formfield': 'uploadedfile',
                        'filename': db,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'fileName': file_short,
                    'cert_name': label,
                    'file_password': password,
                    'file_type': type,
                    'is_getting_val': 'false',
                    'cert_type': 'personal',
                    'uploadType':'iframe'
                },
                json_response=False)

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, label):
    """
    Check if personal cert needs to be uploaded
    """
    ret_obj = search(isvgAppliance, label)

    # Bogus check for now - real test will involve downloading the public personal certificate
    # and compare it with the reference one.
    if len(ret_obj['data']) > 0:
        return True
    else:
        return False
